# -*- coding: utf-8 -*-
"""
    pip_services3_components.config.FileConfigReader
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    File config reader implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.config.ConfigParams import ConfigParams
from .ConfigReader import ConfigReader

class FileConfigReader(ConfigReader):
    """
    Abstract config reader that reads configuration from a file.
    Child classes add support for config files in their specific format
    like JSON, YAML or property files.

    ### Configuration parameters ###

        - path:          path to configuration file
        - parameters:    this entire section is used as template parameters
        - ...
    """
    _path = None

    def __init__(self, path = None):
        """
        Creates a new instance of the config reader.

        :param path: (optional) a path to configuration file.
        """
        super(FileConfigReader, self).__init__()
        self._path = path
        
    def get_path(self):
        """
        Get the path to configuration file.

        :return: the path to configuration file.
        """
        return self._path

    def set_path(self, path):
        """
        Set the path to configuration file.

        :param path: a new path to configuration file.
        """
        self._path = path

    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        super(FileConfigReader, self).configure(config)
        self._path = config.get_as_string_with_default("path", self._path)



# -*- coding: utf-8 -*-
"""
    pip_services3_components.log.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Logger implementations. There exist many different loggers, but all of them are implemented
    differently in various languages. We needed portable classes, that would allow to quickly
    transfer code from one language to another. We can wrap existing loggers into/around our ILogger class.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = [
    'LogLevel', 'LogLevelConverter', 'ILogger', 'Logger', 
    'NullLogger', 'ConsoleLogger', 'CompositeLogger', 
    'LogMessage', 'CachedLogger'
]

from .LogLevel import LogLevel
from .LogLevelConverter import LogLevelConverter
from .ILogger import ILogger
from .Logger import Logger
from .NullLogger import NullLogger
from .ConsoleLogger import ConsoleLogger
from .CompositeLogger import CompositeLogger
from .LogMessage import LogMessage
from .CachedLogger import CachedLogger
from .DefaultLoggerFactory import DefaultLoggerFactory


# -*- coding: utf-8 -*-
"""
    pip_services3_components.auth.DefaultCredentialStoreFactory
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Default credential store factory implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .MemoryCredentialStore import MemoryCredentialStore

from pip_services3_commons.refer.Descriptor import Descriptor
from ..build.Factory import Factory

DefaultCredentialStoreFactoryDescriptor = Descriptor(
    "pip-services", "factory", "credential-store", "default", "1.0"
)

MemoryCredentialStoreDescriptor = Descriptor(
    "pip-services", "credential-store", "memory", "*", "1.0"
)

class DefaultCredentialStoreFactory(Factory):
    """
    Creates ICredentialStore components by their descriptors.
    """
    def __init__(self):
        """
        Create a new instance of the factory.
        """
        self.register_as_type(MemoryCredentialStoreDescriptor, MemoryCredentialStore)



# -*- coding: utf-8 -*-
"""
    pip_services3_components.auth.ICredentialStore
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Credential store interface
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .CredentialParams import CredentialParams

class ICredentialStore:
    """
    Interface for credential stores which are used to store
    and lookup credentials to authenticate against external services.
    """
    def store(self, correlation_id, key, credential):
        """
        Stores credential parameters into the store.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a key to uniquely identify the credential.

        :param credential: a credential to be stored.
        """
        raise NotImplementedError('Method from interface definition')

    def lookup(self, correlation_id, key):
        """
        Lookups credential parameters by its key.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a key to uniquely identify the credential.

        :return: found credential parameters or None if nothing was found
        """
        raise NotImplementedError('Method from interface definition')


# -*- coding: utf-8 -*-
"""
    pip_services3_components.config.CachedConfigReader
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Cached config reader implementation

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pybars import Compiler

from pip_services3_commons.config import IConfigurable
from pip_services3_commons.config import ConfigParams
from .IConfigReader import IConfigReader


class ConfigReader(IConfigReader, IConfigurable):
    """
    Abstract config reader that supports configuration parameterization.

    ### Configuration parameters ###

    parameters:            this entire section is used as template parameters
        - ...
    """
    _parameters = None

    def __init__(self):
        """
        Creates a new instance of the config reader.
        """
        self._parameters = ConfigParams()

    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        parameters = config.get_section("parameters")
        if len(parameters) > 0:
            self._parameters = parameters

    def read_config(self, correlation_id, parameters):
        """
        Reads configuration and parameterize it with given values.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param parameters: values to parameters the configuration or null to skip parameterization.

        :return: ConfigParams configuration.
        """
        raise NotImplementedError('Method is abstract and must be overriden')

    # @staticmethod
    def _parameterize(self, config, parameters):
        """
        Parameterized configuration template given as string with dynamic parameters.

        :param config: a string with configuration template to be parameterized

        :param parameters: dynamic parameters to inject into the template

        :return: a parameterized configuration string.
        """
        parameters = self._parameters.override(parameters)
        compiler = Compiler().compile(config)

        return compiler(parameters)



# -*- coding: utf-8 -*-
"""
    pip_services3_components.connect.ConnectionParams
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Connection parameters implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.config.ConfigParams import ConfigParams
from pip_services3_commons.data import StringValueMap


class ConnectionParams(ConfigParams):
    """
    Contains connection parameters to connect to external services.

    They are used together with credential parameters, but usually stored
    separately from more protected sensitive values.

    ### Configuration parameters ###

        - discovery_key: key to retrieve parameters from discovery service
        - protocol:      connection protocol like http, https, tcp, udp
        - host:          host name or IP address
        - port:          port number
        - uri:           resource URI or connection string with all parameters in it

    In addition to standard parameters ConnectionParams may contain any number of custom parameters

    Example:
        connection = ConnectionParams.from_tuples("protocol", "http",
                                                  "host", "10.1.1.100",
                                                  "port", "8080",
                                                  "cluster", "mycluster")

        host = connection.get_host()                             // Result: "10.1.1.100"
        port = connection.get_port()                             // Result: 8080
        cluster = connection.get_as_nullable_string("cluster")     // Result: "mycluster"
    """

    def __init__(self, map = None):
        """
        Creates a new connection parameters and fills it with values.

        :param map: (optional) an object to be converted into key-value pairs to initialize this connection.
        """
        super(ConnectionParams, self).__init__(map)

    def use_discovery(self):
        """
        Checks if these connection parameters shall be retrieved from [[DiscoveryService]].
        The connection parameters are redirected to [[DiscoveryService]] when discovery_key parameter is set.

        :return: true if connection shall be retrieved from [[DiscoveryService]]
        """
        return "discovery_key" in self

    def get_discovery_key(self):
        """
        Gets the key to retrieve this connection from [[DiscoveryService]].
        If this key is null, than all parameters are already present.

        :return: the discovery key to retrieve connection.
        """
        return self.get_as_nullable_string("discovery_key")

    def set_discovery_key(self, value):
        """
        Sets the key to retrieve these parameters from [[DiscoveryService]].

        :param value: a new key to retrieve connection.
        """
        self.put("discovery_key", value)

    # def get_protocol(self):
    #     """
    #     Gets the connection protocol
    #     Returns: the connection protocol
    #     """
    #     return self.get_as_nullable_string("protocol")

    def get_protocol(self, default_value = None):
        """
        Gets the connection protocol.

        :param default_value: (optional) the default protocol

        :return: the connection protocol or the default value if it's not set.
        """
        return self.get_as_string_with_default("protocol", default_value)

    def set_protocol(self, value):
        """
        Sets the connection protocol.

        :param value: a new connection protocol.
        """
        self.put("protocol", value)

    def get_host(self):
        """
        Gets the host name or IP address.

        :return: the host name or IP address.
        """
        host = self.get_as_nullable_string("host")
        host = host if not (host is None) else self.get_as_nullable_string("ip")
        return host

    def set_host(self, value):
        """
        Sets the host name or IP address.

        :param value: a new host name or IP address.
        """
        self.put("host", value)

    def get_port(self):
        """
        Gets the port number.

        :return: the port number.
        """
        return self.get_as_integer("port")

    def set_port(self, value):
        """
        Sets the port number.

        :param value: a new port number.
        """
        self.set_as_object("port", value)

    def get_uri(self):
        """
        Gets the resource URI or connection string. Usually it includes all connection parameters in it.

        :return: the resource URI or connection string.
        """
        return self.get_as_string("uri")

    def set_uri(self, value):
        """
        Sets the resource URI or connection string.

        :param value: a new resource URI or connection string.
        """
        self.set_as_object("uri", value)

    @staticmethod
    def from_string(line):
        """
        Creates a new ConnectionParams object filled with key-value pairs serialized as a string.

        :param line: a string with serialized key-value pairs as "key1=value1;key2=value2;..."
                     Example: "Key1=123;Key2=ABC;Key3=2016-09-16T00:00:00.00Z"

        :return: a new ConnectionParams object.
        """
        map = StringValueMap.from_string(line)
        return ConnectionParams(map)

    @staticmethod
    def from_tuples(*tuples):
        """
        Creates a new ConnectionParams object filled with provided key-value pairs called tuples.
        Tuples parameters contain a sequence of key1, value1, key2, value2, ... pairs.

        :param tuples: the tuples to fill a new ConnectionParams object.

        :return: a new ConnectionParams object.
        """
        map = StringValueMap.from_tuples_array(tuples)
        return ConnectionParams(map)

    @staticmethod
    def many_from_config(config):
        """
        Retrieves all ConnectionParams from configuration parameters
        from "connections" section. If "connection" section is present instead,
        than it returns a list with only one ConnectionParams.

        :param config: a configuration parameters to retrieve connections

        :return: a list of retrieved ConnectionParams
        """
        result = []

        # Try to get multiple connections first
        connections = config.get_section("connections")
        if len(connections) > 0:
            sections_names = connections.get_section_names()
            for section in sections_names:
                connection = connections.get_section(section)
                result.append(ConnectionParams(connection))
        # Then try to get a single connection
        else:
            connection = config.get_section("connection")
            result.append(ConnectionParams(connection))

        return result

    @staticmethod
    def from_config(config):
        """
        Retrieves a single ConnectionParams from configuration parameters
        from "connection" section. If "connections" section is present instead,
        then is returns only the first connection element.

        :param config: ConnectionParams, containing a section named "connection(s)".

        :return: the generated ConnectionParams object.
        """
        connections = ConnectionParams.many_from_config(config)
        return connections[0] if len(connections) > 0 else None



# -*- coding: utf-8 -*-
"""
    pip_services3_components.auth.MemoryCredentialStore
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Memory credential store implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.config.ConfigParams import ConfigParams
from pip_services3_commons.config.IReconfigurable import IReconfigurable
from pip_services3_commons.data.StringValueMap import StringValueMap
from .ICredentialStore import ICredentialStore
from .CredentialParams import CredentialParams

class MemoryCredentialStore(ICredentialStore, IReconfigurable):
    """
    Credential store that keeps credentials in memory.

    ### Configuration parameters ###

        - [credential key 1]:
        - ...                          credential parameters for key 1
        - [credential key 2]:
        - ...                          credential parameters for key N
        - ...

    Example:
        config = ConfigParams.from_tuples("key1.user", "jdoe",
                                          "key1.pass", "pass123",
                                          "key2.user", "bsmith",
                                          "key2.pass", "mypass")

        credentialStore = MemoryCredentialStore()
        credentialStore.read_credentials(config)
        credentialStore.lookup("123", "key1")
    """
    _items = None

    def __init__(self, credentials = None):
        """
        Creates a new instance of the credential store.

        :param credentials: (optional) configuration with credential parameters.
        """
        self._items = StringValueMap()
        if not (credentials is None):
            self.configure(credentials)

    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        self.read_credentials(config)

    def read_credentials(self, credentials):
        """
        Reads credentials from configuration parameters.
        Each section represents an individual CredentialParams

        :param credentials: configuration parameters to be read
        """
        self._items.clear()
        for key in credentials.get_key_names():
            value = credentials.get_as_nullable_string(key)
            self._items.append(CredentialParams.from_tuples([key, value]))

    def store(self, correlation_id, key, credential):
        """
        Stores credential parameters into the store.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a key to uniquely identify the credential parameters.

        :param credential: a credential parameters to be stored.
        """
        if not (credential is None):
            self._items.put(key, credential)
        else:
            self._items.remove(key)

    def lookup(self, correlation_id, key):
        """
        Lookups credential parameters by its key.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a key to uniquely identify the credential.

        :return: found credential parameters or None if nothing was found
        """
        credential = self._items.get_as_object(key)



# -*- coding: utf-8 -*-
"""
    pip_services3_components.config.IConfigReader
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    interface for configuration readers
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

class IConfigReader:
    """
    Interface for configuration readers that retrieve configuration from various sources
    and make it available for other components.

    Some IConfigReader implementations may support configuration parameterization.
    The parameterization allows to use configuration as a template and inject there dynamic values.
    The values may come from application command like arguments or environment variables.
    """
    def read_config(self, correlation_id, parameters):
        """
        Reads configuration and parameterize it with given values.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param parameters: values to parameters the configuration or null to skip parameterization.

        :return: ConfigParams configuration.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services3_components.counters.ITimingCallback
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for performance timing callbacks.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

class ITimingCallback:
    """
    Interface for a callback to end measurement of execution elapsed time.
    """

    def end_timing(self, name, elapsed):
        """
        Ends measurement of execution elapsed time and updates specified counter.

        :param name: a counter name

        :param elapsed: execution elapsed time in milliseconds to update the counter.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services3_components.log.CachedLogger
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Cached logger implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import time
import threading
import socket

from .ILogger import ILogger
from .Logger import Logger
from .LogMessage import LogMessage
from pip_services3_commons.errors.ErrorDescriptionFactory import ErrorDescriptionFactory
from pip_services3_commons.config.IReconfigurable import IReconfigurable

class CachedLogger(Logger, IReconfigurable):
    """
    Abstract logger that caches captured log messages in memory and periodically dumps them.
    Child classes implement saving cached messages to their specified destinations.

    ### Configuration parameters ###

        - level:             maximum log level to capture
        - source:            source (context) name
        - options:
            - interval:        interval in milliseconds to save log messages (default: 10 seconds)
            - max_cache_size:  maximum number of messages stored in this cache (default: 100)

    ### References ###
        - *:context-info:*:*:1.0     (optional) [[ContextInfo]] to detect the context id and specify counters source
    """
    _cache = None
    _updated = None
    _last_dump_time = None
    _interval = 60000
    _lock = None


    def __init__(self):
        """
        Creates a new instance of the logger.
        """
        self._cache = []
        self._updated = False
        self._last_dump_time = time.perf_counter() * 1000
        self._lock = threading.Lock()


    def _write(self, level, correlation_id, ex, message):
        """
        Writes a log message to the logger destination.

        :param level: a log level.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param ex: an error object associated with this message.

        :param message: a human-readable message to log.
        """
        error = ErrorDescriptionFactory.create(ex) if not (ex is None) else None
        source = socket.gethostname() # Todo: add process/module name
        log_message = LogMessage(level, source, correlation_id, error, message)
        
        self._lock.acquire()
        try:
            self._cache.append(log_message)
        finally:
            self._lock.release()

        self._update()


    def _save(self, messages):
        """
        Saves log messages from the cache.

        :param messages: a list with log messages
        """
        raise NotImplementedError('Method from abstract implementation')


    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        self._interval = config.get_as_float_with_default("interval", self._interval)


    def clear(self):
        """
        Clears (removes) all cached log messages.
        """
        self._lock.acquire()
        try:
            self._cache = []
            self._updated = False
        finally:
            self._lock.release()


    def dump(self):
        """
        Dumps (writes) the currently cached log messages.
        """
        if self._updated:
            self._lock.acquire()
            try:
                if not self._updated:
                    return
                
                messages = self._cache
                self._cache = []
                
                self._save(messages)

                self._updated = False
                self._last_dump_time = time.perf_counter() * 1000
            finally:
                self._lock.release()


    def _update(self):
        """
        Makes message cache as updated and dumps it when timeout expires.
        """
        self._updated = True
        
        if time.perf_counter() * 1000 > self._last_dump_time + self._interval:
            try:
                self.dump()
            except:
                # Todo: decide what to do
                pass



# -*- coding: utf-8 -*-
"""
    pip_services3_components.counters.Counter
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Counter object implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

class Counter(object):
    """
    Data object to store measurement for a performance counter.
    This object is used by [[CachedCounters]] to store counters.
    """
    name = None
    type = None
    last = None
    count = None
    min = None
    max = None
    average = None
    time = None

    def __init__(self, name= None, tipe = None):
        """
        Creates a instance of the data obejct

        :param name: a counter name.

        :param tipe: a counter type.
        """
        self.name = name
        self.type = tipe



# -*- coding: utf-8 -*-
"""
    pip_services3_components.cache.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Abstract implementation of various distributed caches. We can save an object
    to cache and retrieve it object by its key, using various implementations.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = [
    'ICache', 'CacheEntry', 'NullCache',
    'MemoryCache', 'DefaultCacheFactory'
]

from .ICache import ICache
from .CacheEntry import CacheEntry
from .NullCache import NullCache
from .MemoryCache import MemoryCache
from .DefaultCacheFactory import DefaultCacheFactory


# -*- coding: utf-8 -*-
"""
    pip_services3_components.count.NullCounter
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Null counter implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .ICounters import ICounters
from .Timing import Timing

class NullCounters(ICounters):
    """
    Dummy implementation of performance counters that doesn't do anything.
    It can be used in testing or in situations when counters is required but shall be disabled.
    """

    def begin_timing(self, name):
        """
        Begins measurement of execution time interval.
        It returns [[Timing]] object which has to be called at
        [[Timing.endTiming]] to end the measurement and update the counter.

        :param name: a counter name of Interval type.

        :return: a [[Timing]] callback object to end timing.
        """
        return Timing()

    def stats(self, name, value):
        """
        Calculates min/average/max statistics based on the current and previous values.

        :param name: a counter name of Statistics type

        :param value: a value to update statistics
        """
        pass

    def last(self, name, value):
        """
        Records the last calculated measurement value.
        Usually this method is used by metrics calculated externally.

        :param name: a counter name of Last type.

        :param value: a last value to record.
        """
        pass

    def timestamp_now(self, name):
        """
        Records the current time as a timestamp.

        :param name: a counter name of Timestamp type.
        """
        pass

    def timestamp(self, name, value):
        """
        Records the given timestamp.

        :param name: a counter name of Timestamp type.

        :param value: a timestamp to record.
        """
        pass

    def increment_one(self, name):
        """
        Increments counter by 1.

        :param name: a counter name of Increment type.
        """
        pass

    def increment(self, name, value):
        """
        Increments counter by given value.

        :param name: a counter name of Increment type.

        :param value: a value to add to the counter.
        """
        pass



# -*- coding: utf-8 -*-
"""
    pip_services3_components.counters.CachedCounters
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Cached counters implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import time
import threading
import datetime

from .ICounters import ICounters
from .Counter import Counter
from .CounterType import CounterType
from .ITimingCallback import ITimingCallback
from .Timing import Timing
from pip_services3_commons.config.IReconfigurable import IReconfigurable

class CachedCounters(ICounters, IReconfigurable, ITimingCallback):
    """
    Abstract implementation of performance counters that measures and stores counters in memory.
    Child classes implement saving of the counters into various destinations.

    ### Configuration parameters ###

        - options:
            - interval:        interval in milliseconds to save current counters measurements (default: 5 mins)
            - reset_timeout:   timeout in milliseconds to reset the counters. 0 disables the reset (default: 0)
    """
    _default_interval = 300000

    _cache = None
    _updated = None
    _last_dump_time = None
    _interval = None
    _lock = None


    def __init__(self):
        """
        Creates a new CachedCounters object.
        """
        self._cache = {}
        self._updated = False
        self._last_dump_time = time.perf_counter()
        self._interval = self._default_interval
        self._lock = threading.Lock()


    def _save(self, counters):
        """
        Saves the current counters measurements.

        :param counters: current counters measurements to be saves.
        """
        raise NotImplementedError('Method from abstract implementation')


    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        self._interval = config.get_as_float_with_default("interval", self._interval)


    def clear(self, name):
        """
        Clears (resets) a counter specified by its name.

        :param name: a counter name to clear.
        """
        self._lock.acquire()
        try:
            del self._cache[name]
        finally:
            self._lock.release()


    def clear_all(self):
        """
        Clears (resets) all counters.
        """
        self._lock.acquire()
        try:
            self._cache = {}
            self._updated = False
        finally:
            self._lock.release()


    def dump(self):
        """
        Dumps (saves) the current values of counters.
        """
        if self._updated:
            messages = self.get_all()
            self._save(messages)

            self._lock.acquire()
            try:
                self._updated = False
                current_time = time.perf_counter() * 1000
                self._last_dump_time = current_time
            finally:
                self._lock.release()


    def _update(self):
        """
        Makes counter measurements as updated and dumps them when timeout expires.
        """
        self._updated = True
        
        current_time = time.perf_counter() * 1000
        if current_time > self._last_dump_time + self._interval:
            try:
                self.dump()
            except:
                # Todo: decide what to do
                pass


    def get_all(self):
        """
        Gets all captured counters.

        :return: a list with counters.
        """
        self._lock.acquire()
        try:
            return list(self._cache.values())
        finally:
            self._lock.release()


    def get(self, name, typ):
        """
        Gets a counter specified by its name.
        It counter does not exist or its type doesn't match the specified type
        it creates a new one.

        :param name: a counter name to retrieve.

        :param typ: a counter type.

        :return: an existing or newly created counter of the specified type.
        """
        if name is None or len(name) == 0:
            raise Exception("Counter name was not set")

        self._lock.acquire()
        try:
            counter = self._cache[name] if name in self._cache else None

            if counter is None or counter.type != typ:
                counter = Counter(name, typ)
                self._cache[name] = counter

            return counter
        finally:
            self._lock.release()


    def _calculate_stats(self, counter, value):
        if counter is None:
            raise Exception("Missing counter")

        counter.last = value
        counter.count = counter.count + 1 if not (counter.count is None) else 1
        counter.max = max(counter.max, value) if not (counter.max is None) else value
        counter.min = min(counter.min, value) if not (counter.min is None) else value
        counter.average = (float(counter.average * (counter.count - 1)) + value) / counter.count \
            if not (counter.average is None) and counter.count > 0 else value


    def begin_timing(self, name):
        """
        Begins measurement of execution time interval.
        It returns [[Timing]] object which has to be called at
        [[Timing.endTiming]] to end the measurement and update the counter.

        :param name: a counter name of Interval type.

        :return: a [[Timing]] callback object to end timing.
        """
        return Timing(name, self)

    def end_timing(self, name, elapsed):
        """
        Ends measurement of execution elapsed time and updates specified counter.

        :param name: a counter name

        :param elapsed: execution elapsed time in milliseconds to update the counter.
        """
        counter = self.get(name, CounterType.Interval)
        self._calculate_stats(counter, elapsed)
        self._update()


    def stats(self, name, value):
        """
        Calculates min/average/max statistics based on the current and previous values.

        :param name: a counter name of Statistics type

        :param value: a value to update statistics
        """
        counter = self.get(name, CounterType.Statistics)
        self._calculate_stats(counter, value)
        self._update()


    def last(self, name, value):
        """
        Records the last calculated measurement value.
        Usually this method is used by metrics calculated externally.

        :param name: a counter name of Last type.

        :param value: a last value to record.
        """
        counter = self.get(name, CounterType.LastValue)
        counter.last = value
        self._update()


    def timestamp_now(self, name):
        """
        Records the current time as a timestamp.

        :param name: a counter name of Timestamp type.
        """
        self.timestamp(name, datetime.datetime.utcnow())


    def timestamp(self, name, value):
        """
        Records the given timestamp.

        :param name: a counter name of Timestamp type.

        :param value: a timestamp to record.
        """
        counter = self.get(name, CounterType.Timestamp)
        counter.time = value if not (value is None) else datetime.datetime.utcnow()
        self._update()


    def increment_one(self, name):
        """
        Increments counter by 1.

        :param name: a counter name of Increment type.
        """
        self.increment(name, 1)


    def increment(self, name, value):
        """
        Increments counter by given value.

        :param name: a counter name of Increment type.

        :param value: a value to add to the counter.
        """
        counter = self.get(name, CounterType.Increment)
        counter.count = counter.count + value if not (counter.count is None) else value
        self._update()



# -*- coding: utf-8 -*-
"""
    pip_services3_components.Component
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Component implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.config import IConfigurable
from .log.CompositeLogger import CompositeLogger
from .count.CompositeCounters import CompositeCounters
from pip_services3_commons.refer import IReferenceable
from pip_services3_commons.refer import DependencyResolver

class Component(IConfigurable, IReferenceable):
    """
    Component class implementation.
    Abstract component that supportes configurable dependencies,
    logging and performance counters.

    ### Configuration parameters ###

        - dependencies:
            - [dependency name 1]: Dependency 1 locator (descriptor)
            - ...
            - [dependency name N]: Dependency N locator (descriptor)

    ### References ###
        - *:counters:*:*:1.0     (optional) [[ICounters]] components to pass collected measurements
        - *:logger:*:*:1.0       (optional) [[ILogger]] components to pass log messages
        - ...                                    References must match configured dependencies.
    """
    _logger = None
    _counters = None
    _dependency_resolver = None

    def __init__(self):
        self._logger = CompositeLogger()
        self._counters = CompositeCounters()
        self._dependency_resolver = DependencyResolver()

    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        self._dependency_resolver.configure(config)
        self._logger.configure(config)

    def set_references(self, references):
        """
        Sets references to dependent components.

        :param references: references to locate the component dependencies.
        """
        self._dependency_resolver.set_references(references)
        self._logger.set_references(references)
        self._counters.set_references(references)



# -*- coding: utf-8 -*-
"""
    pip_services3_components.log.DefaultLoggerFactory
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Default logger factory implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .NullLogger import NullLogger
from .ConsoleLogger import ConsoleLogger
from .CompositeLogger import CompositeLogger

from pip_services3_commons.refer.Descriptor import Descriptor
from ..build.Factory import Factory

DefaultLoggerFactoryDescriptor = Descriptor(
    "pip-services", "factory", "logger", "default", "1.0"
)

NullLoggerDescriptor = Descriptor(
    "pip-services", "logger", "null", "*", "1.0"
)

ConsoleLoggerDescriptor = Descriptor(
    "pip-services", "logger", "console", "*", "1.0"
)

CompositeLoggerDescriptor = Descriptor(
    "pip-services", "logger", "composite", "*", "1.0"
)

class DefaultLoggerFactory(Factory):
    """
    Creates [[ILogger]] components by their descriptors.
    """
    def __init__(self):
        """
        Create a new instance of the factory.
        """
        self.register_as_type(NullLoggerDescriptor, NullLogger)
        self.register_as_type(ConsoleLoggerDescriptor, ConsoleLogger)
        self.register_as_type(CompositeLoggerDescriptor, CompositeLogger)



# -*- coding: utf-8 -*-
"""
    pip_services3_components.info.DefaultInfoFactory
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Default info factory implementation

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.refer.Descriptor import Descriptor
from ..build.Factory import Factory
from .ContextInfo import ContextInfo


ContextInfoDescriptor = Descriptor("pip-services", "context-info", "default", "*", "1.0")

ContainerInfoDescriptor = Descriptor("pip-services", "container-info", "default", "*", "1.0")

class DefaultInfoFactory(Factory):
    """
    Creates information components by their descriptors.
    """
    def __init__(self):
        """
        Create a new instance of the factory.
        """
        self.register_as_type(ContextInfoDescriptor, ContextInfo)
        self.register_as_type(ContainerInfoDescriptor, ContextInfo)


# -*- coding: utf-8 -*-
"""
    pip_services3_components.auth.CredentialResolver
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Credential resolver implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .CredentialParams import CredentialParams
from pip_services3_commons.refer.Descriptor import Descriptor
from .ICredentialStore import ICredentialStore
from pip_services3_commons.config.IConfigurable import IConfigurable
from pip_services3_commons.refer.IReferenceable import IReferenceable
from pip_services3_commons.refer.ReferenceException import ReferenceException
from pip_services3_commons.errors.ApplicationException import ApplicationException

class CredentialResolver(IConfigurable, IReferenceable):
    """
    Helper class to retrieve component credentials.

    If credentials are configured to be retrieved from
    [[ICredentialStore]], it automatically locates [[ICredentialStore]]
    in component references and retrieve credentials from there using store_key parameter.

    ### Configuration parameters ###

    credential:
        - store_key:                   (optional) a key to retrieve the credentials from [[ICredentialStore]]
        - ...                          other credential parameters

    credentials:                   alternative to credential
        - [credential params 1]:       first credential parameters
        - ...                      credential parameters for key 1
        - ...
        - [credential params N]:       Nth credential parameters
        - ...                      credential parameters for key N

    ### References ###
        - *:credential-store:*:*:1.0  (optional) Credential stores to resolve credentials

    Example:
          config = ConfigParams.from_tuples("credential.user", "jdoe",
                                            "credential.pass",  "pass123")

          credentialResolver = CredentialResolver()
          credentialResolver.configure(config)
          credentialResolver.set_references(references)
          credentialResolver.lookup("123")

    """
    _credentials = None
    _references = None

    def __init__(self, config = None, references = None):
        """
        Creates a new instance of credentials resolver.

        :param config: (optional) component configuration parameters

        :param references: (optional) component references
        """
        self._credentials = []
        if not (config is None):
            self.configure(config)
        if not (references is None):
            self.set_references(references)

    def set_references(self, references):
        """
        Sets references to dependent components.

        :param references: references to locate the component dependencies.
        """
        self._references = references

    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        credentials = CredentialParams.many_from_config(config)
        for credential in credentials:
            self._credentials.append(credential)

    def get_all(self):
        """
        Gets all credentials configured in component configuration.

        Redirect to CredentialStores is not done at this point.
        If you need fully fleshed credential use [[lookup]] method instead.

        :return: a list with credential parameters
        """
        return list(self._credentials)

    def add(self, connection):
        """
        Adds a new credential to component credentials

        :param connection: new credential parameters to be added
        """
        self._credentials.append(connection)

    def _lookup_in_stores(self, correlation_id, credential):
        if credential.use_credential_store() == False: return None
        
        key = credential.get_store_key()
        if self._references is None:
            return None
        
        descriptor = Descriptor("*", "credential_store", "*", "*", "*")
        components = self._references.get_optional(descriptor)
        if len(components) == 0:
            raise ReferenceException(correlation_id, "Credential store wasn't found to make lookup")
        
        # TODO: create assync 
        for component in components:
            if isinstance(component, ICredentialStore):
                resolved_credential = component.lookup(correlation_id, key)
                if not (resolved_credential is None):
                    return resolved_credential

        return None

    def lookup(self, correlation_id):
        """
        Looks up component credential parameters. If credentials are configured to be retrieved
        from Credential store it finds a [[ICredentialStore]] and lookups credentials there.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :return: resolved credential parameters or None if nothing was found.
        """
        if len(self._credentials) == 0: return None
        
        # Return connection that doesn't require discovery
        for credential in self._credentials:
            if not credential.use_credential_store():
                return credential
        
        # Return connection that require discovery
        for credential in self._credentials:
            if credential.use_credential_store():
                resolved_connection = self._lookup_in_stores(correlation_id, credential)
                if not (resolved_connection is None):
                    return resolved_connection
        
        return None




# -*- coding: utf-8 -*-
"""
    pip_services3_components.auth.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Contains credentials implementation.

    Credentials – passwords, logins,
    application keys, secrets. This information is usually linked with connection parameters.
    Connection parameters separate from authentication, because auth is saved as a secret,
    and stored separately from configuration parameters (host name, ip addresses).
    They need added security and protection, so they were separated.

    Credential parameters include various credentials.

    Interfaces and abstract classes for credential stores, which can save or retrieve various credential parameters.

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = ['CredentialParams', 'ICredentialStore', 'CredentialResolver',
    'MemoryCredentialStore', 'DefaultCredentialStoreFactory']

from .CredentialParams import CredentialParams
from .ICredentialStore import ICredentialStore
from .CredentialResolver import CredentialResolver
from .MemoryCredentialStore import MemoryCredentialStore
from .DefaultCredentialStoreFactory import DefaultCredentialStoreFactory




# -*- coding: utf-8 -*-
"""
    pip_services3_components.log.CompositeCounters
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Composite counters implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .ICounters import ICounters
from .ITimingCallback import ITimingCallback
from .Timing import Timing
from pip_services3_commons.refer.Descriptor import Descriptor
from pip_services3_commons.refer.IReferenceable import IReferenceable

class CompositeCounters(ICounters, ITimingCallback, IReferenceable):
    """
    Aggregates all counters from component references under a single component.

    It allows to capture metrics and conveniently send them to multiple destinations.

    ### References ###

        - *:counters:*:*:1.0     (optional) ICounters components to pass collected measurements

    Example:
        class MyComponent(IReferenceable):
            _counters = CompositeCounters()

        def set_references(self, references):
            self._counters.set_references(references)

        def my_method(self):
            self._counters.increment("mycomponent.mymethod.calls")
            timing = this._counters.beginTiming("mycomponent.mymethod.exec_time")
            ...

            timing.endTiming()
    """
    _counters = None

    def __init__(self, references = None):
        """
        Creates a new instance of the counters.

        :param references: references to locate the component dependencies.
        """
        self._counters = []

        if not (references is None):
            self.set_references(references)
            
    def set_references(self, references):
        """
        Sets references to dependent components.

        :param references: references to locate the component dependencies.
        """
        descriptor = Descriptor(None, "counters", None, None, None)
        counters = references.get_optional(descriptor)
        for counter in counters:
            if isinstance(counter, ICounters):
                self._counters.append(counter)

    def begin_timing(self, name):
        """
        Begins measurement of execution time interval.
        It returns [[Timing]] object which has to be called at
        [[Timing.endTiming]] to end the measurement and update the counter.

        :param name: a counter name of Interval type.

        :return: a [[Timing]] callback object to end timing.
        """
        return Timing(name, self)

    def end_timing(self, name, elapsed):
        """
        Ends measurement of execution elapsed time and updates specified counter.

        :param name: a counter name

        :param elapsed: execution elapsed time in milliseconds to update the counter.
        """
        for counter in self._counters:
            if isinstance(counter, ITimingCallback):
                counter.end_timing(name, elapsed)

    def stats(self, name, value):
        """
        Calculates min/average/max statistics based on the current and previous values.

        :param name: a counter name of Statistics type

        :param value: a value to update statistics
        """
        for counter in self._counters:
            counter.stats(name, value)

    def last(self, name, value):
        """
        Records the last calculated measurement value.
        Usually this method is used by metrics calculated externally.

        :param name: a counter name of Last type.

        :param value: a last value to record.
        """
        for counter in self._counters:
            counter.last(name, value)

    def timestamp_now(self, name):
        """
        Records the current time as a timestamp.

        :param name: a counter name of Timestamp type.
        """
        for counter in self._counters:
            counter.timestamp_now(name)

    def timestamp(self, name, value):
        """
        Records the given timestamp.

        :param name: a counter name of Timestamp type.

        :param value: a timestamp to record.
        """
        for counter in self._counters:
            counter.timestamp(name, value)

    def increment_one(self, name):
        """
        Increments counter by 1.

        :param name: a counter name of Increment type.
        """
        for counter in self._counters:
            counter.increment_one(name)

    def increment(self, name, value):
        """
        Increments counter by given value.

        :param name: a counter name of Increment type.

        :param value: a value to add to the counter.
        """
        for counter in self._counters:
            counter.increment(name, value)



# -*- coding: utf-8 -*-
"""
    pip_services3_components.counters.CounterType
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Counter type enumeration
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

class CounterType(object):
    """
    Types of counters that measure different types of metrics
    """
    Interval = 0
    LastValue = 1
    Statistics = 2
    Timestamp = 3
    Increment = 4

    @staticmethod
    def to_string(typ):
        if typ == CounterType.Interval:
            return "INTERVAL" 
        if typ == CounterType.LastValue:
            return "LAST" 
        if typ == CounterType.Statistics:
            return "STATS" 
        if typ == CounterType.Timestamp:
            return "TIME"
        if typ == CounterType.Increment:
            return "COUNT"
        return "UNDEF"



# -*- coding: utf-8 -*-
"""
    pip_services3_components.count.DefaultCountersFactory
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Default counters factory implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .NullCounters import NullCounters
from .LogCounters import LogCounters
from .CompositeCounters import CompositeCounters

from pip_services3_commons.refer.Descriptor import Descriptor
from ..build.Factory import Factory

DefaultCountersFactoryDescriptor = Descriptor(
    "pip-services", "factory", "counters", "default", "1.0"
)

NullCountersDescriptor = Descriptor(
    "pip-services", "counters", "null", "*", "1.0"
)

LogCountersDescriptor = Descriptor(
    "pip-services", "counters", "log", "*", "1.0"
)

CompositeCountersDescriptor = Descriptor(
    "pip-services", "counters", "composite", "*", "1.0"
)

class DefaultCountersFactory(Factory):
    """
    Creates [[ICounters]] components by their descriptors.
    """
    def __init__(self):
        """
        Create a new instance of the factory.
        """
        self.register_as_type(NullCountersDescriptor, NullCounters)
        self.register_as_type(LogCountersDescriptor, LogCounters)
        self.register_as_type(CompositeCountersDescriptor, CompositeCounters)



# -*- coding: utf-8 -*-
"""
    pip_services3_components.counters.ICounters
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for performance counters components.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

class ICounters:
    """
    Interface for performance counters that measure execution metrics.
    The performance counters measure how code is performing:
    how fast or slow, how many transactions performed, how many objects
    are stored, what was the latest transaction time and so on.

    They are critical to monitor and improve performance, scalability and reliability of code in production.
    """

    def begin_timing(self, name):
        """
        Begins measurement of execution time interval.
        It returns [[Timing]] object which has to be called at
        [[Timing.endTiming]] to end the measurement and update the counter.

        :param name: a counter name of Interval type.

        :return: a [[Timing]] callback object to end timing.
        """
        raise NotImplementedError('Method from interface definition')

    def stats(self, name, value):
        """
        Calculates min/average/max statistics based on the current and previous values.

        :param name: a counter name of Statistics type

        :param value: a value to update statistics
        """
        raise NotImplementedError('Method from interface definition')

    def last(self, name, value):
        """
        Records the last calculated measurement value.
        Usually this method is used by metrics calculated externally.

        :param name: a counter name of Last type.

        :param value: a last value to record.
        """
        raise NotImplementedError('Method from interface definition')

    def timestamp_now(self, name):
        """
        Records the current time as a timestamp.

        :param name: a counter name of Timestamp type.
        """
        raise NotImplementedError('Method from interface definition')

    def timestamp(self, name, value):
        """
        Records the given timestamp.

        :param name: a counter name of Timestamp type.

        :param value: a timestamp to record.
        """
        raise NotImplementedError('Method from interface definition')

    def increment_one(self, name):
        """
        Increments counter by 1.

        :param name: a counter name of Increment type.
        """
        raise NotImplementedError('Method from interface definition')

    def increment(self, name, value):
        """
        Increments counter by given value.

        :param name: a counter name of Increment type.

        :param value: a value to add to the counter.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services3_components.log.Logger
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Abstract logger implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from abc import ABC, abstractmethod

from pip_services3_commons.refer import IReferenceable, Descriptor

from .LogLevel import LogLevel
from .ILogger import ILogger
from .LogLevelConverter import LogLevelConverter
from pip_services3_commons.config.IReconfigurable import IReconfigurable


class Logger(ILogger, IReconfigurable, IReferenceable, ABC):
    """
    Abstract logger that captures and formats log messages.
    Child classes take the captured messages and write them to their specific destinations.

    ### Configuration parameters ###

    Parameters to pass to the [[configure]] method for component configuration:
        - level:             maximum log level to capture
        - source:            source (context) name

    ### References ###
        - *:context-info:*:*:1.0     (optional) [[ContextInfo]] to detect the context id and specify counters source
    """
    _level = LogLevel.Info
    _source = None

    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        self._level = LogLevelConverter.to_log_level(config.get_as_object("level"))
        self._source = config.get_as_string_with_default("source", self._source)

    def set_references(self, references):
        """
        Sets references to dependent components.

        :param references: references to locate the component dependencies.
        """
        context_info = references.get_one_optional(Descriptor("pip-services", "context-info", "*", "*", "1.0"))
        if context_info is not None and self._source is None:
            self._source = context_info.name

    def get_level(self):
        """
        Gets the maximum log level. Messages with higher log level are filtered out.

        :return: the maximum log level.
        """
        return self._level

    def set_level(self, level):
        """
        Set the maximum log level.

        :param level: a new maximum log level.
        """
        self._level = level

    @abstractmethod
    def _write(self, level, correlation_id, error, message):
        """
        Writes a log message to the logger destination.

        :param level: a log level.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param error: an error object associated with this message.

        :param message: a human-readable message to log.
        """
        raise NotImplementedError('Method from abstract implementation')

    def _format_and_write(self, level, correlation_id, error, message, *args, **kwargs):
        """
        Formats the log message and writes it to the logger destination.

        :param level: a log level.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param error: an error object associated with this message.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        if not (message is None) and len(message) > 0 and len(kwargs) > 0:
            message = message.format(*args, **kwargs)
        self._write(level, correlation_id, error, message)

    def log(self, level, correlation_id, error, message, *args, **kwargs):
        """
        Logs a message at specified log level.

        :param level: a log level.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param error: an error object associated with this message.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        self._format_and_write(level, correlation_id, error, message, args, kwargs)

    def fatal(self, correlation_id, error, message, *args, **kwargs):
        """
        Logs fatal (unrecoverable) message that caused the process to crash.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param error: an error object associated with this message.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        self._format_and_write(LogLevel.Fatal, correlation_id, error, message, args, kwargs)

    def error(self, correlation_id, error, message, *args, **kwargs):
        """
        Logs recoverable application error.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param error: an error object associated with this message.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        self._format_and_write(LogLevel.Error, correlation_id, error, message, args, kwargs)

    def warn(self, correlation_id, message, *args, **kwargs):
        """
        Logs a warning that may or may not have a negative impact.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        self._format_and_write(LogLevel.Warn, correlation_id, None, message, args, kwargs)

    def info(self, correlation_id, message, *args, **kwargs):
        """
        Logs an important information message

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        self._format_and_write(LogLevel.Info, correlation_id, None, message, args, kwargs)

    def debug(self, correlation_id, message, *args, **kwargs):
        """
        Logs a high-level debug information for troubleshooting.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        self._format_and_write(LogLevel.Debug, correlation_id, None, message, args, kwargs)

    def trace(self, correlation_id, message, *args, **kwargs):
        """
        Logs a low-level debug information for troubleshooting.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        self._format_and_write(LogLevel.Trace, correlation_id, None, message, args, kwargs)

    def _compose_error(self, error):
        """
        Composes an human-readable error description

        :param error: an error to format.
        :return: a human-reable error description.
        """
        builder = ''

        builder += error.message

        app_error = error
        if app_error.cause:
            builder += ' Cause by: '
            builder += str(app_error.cause)

        if error.stack_trace:
            builder += ' Stack trace '
            builder += error.stack_trace

        return builder



# -*- coding: utf-8 -*-
"""
    pip_services3_components.connect.DefaultConfigReaderFactory
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Default discovery factory implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .MemoryConfigReader import MemoryConfigReader
from .JsonConfigReader import JsonConfigReader
from .YamlConfigReader import YamlConfigReader

from pip_services3_commons.refer.Descriptor import Descriptor
from ..build.Factory import Factory

DefaultDiscoveryFactoryDescriptor = Descriptor(
    "pip-services", "factory", "config-reader", "default", "1.0"
)

MemoryConfigReaderDescriptor = Descriptor(
    "pip-services", "config-reader", "memory", "*", "1.0"
)

JsonConfigReaderDescriptor = Descriptor(
    "pip-services", "config-reader", "json", "*", "1.0"
)

YamlConfigReaderDescriptor = Descriptor(
    "pip-services", "config-reader", "yaml", "*", "1.0"
)

class DefaultConfigReaderFactory(Factory):
    """
    Creates [[IConfigReader]] components by their descriptors.
    """
    def __init__(self):
        """
        Create a new instance of the factory.
        """
        self.register_as_type(MemoryConfigReaderDescriptor, MemoryConfigReader)
        self.register_as_type(JsonConfigReaderDescriptor, JsonConfigReader)
        self.register_as_type(YamlConfigReaderDescriptor, YamlConfigReader)



# -*- coding: utf-8 -*-
"""
    pip_services3_components.cache.ICache
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for caching components.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

class ICache:
    """
    Interface for caches that are used to cache values to improve performance.
    """

    def retrieve(self, correlation_id, key):
        """
        Retrieves cached value from the cache using its key.
        If value is missing in the cache or expired it returns None.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a unique value key.

        :return: a cached value or None if value wasn't found or timeout expired.
        """
        raise NotImplementedError('Method from interface definition')

    def store(self, correlation_id, key, value, timeout):
        """
        Stores value in the cache with expiration time.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a unique value key.

        :param value: a value to store.

        :param timeout: expiration timeout in milliseconds.

        :return: a cached value stored in the cache.
        """
        raise NotImplementedError('Method from interface definition')

    def remove(self, correlation_id, key):
        """
        Removes a value from the cache by its key.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a unique value key.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services3_components.connect.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Contains implementation of connection parameters, using various connection strings,
    which are stripped of all credentials. If we need to configure a service, the port,
    ip address, protocol, and other parameters – we use the ConnectionParams object, and
    relevant helper classes (like [[ConnectionResolver]]), for acquiring these parameters,
    and for discovery of objects, components (which store and retrieve connection parameters).

    ### Discovery ###

    Service that store a registry of various end-points (what services are where, and how to
    connect to them). It knows the end-points, but doesn't have the credentials to connect to them.
    Separated for security reasons. IDiscovery – interface for creating registries.
    MemoryDiscovery – registry that is stored in memory.

    There exist 2 types of discovery:
        - Static discovery: all services have static IP addresses (like DNS, which also works using static
        discovery) that are configured from the start and don't change along the way. As of lately, used
        more often than dynamic, because it is simpler to use and more reliable.
        - Proxy (or reverse proxy) is created with a dns name, and all the dynamics of
        starting/restarting/switching from one host to another – everything is nice and clear for the clients.
        Infrastructure does all the hard work out of the box.
        - Configure sets the static registry.
        - Dynamic discovery: every time a service starts, it registers its address in the discovery service
        ("Service name" at the following address "IP"). Clients then ask to resolve the address
        by which the requested service can be reached. The service has a general name, by which other
        services can resolve it.
        - If a service stops working, you need to refresh its address, clean stale addresses,
        heartbeats must be used – lots of problems and challenges. One service can have more than one address.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = [ 'ConnectionParams', 'IDiscovery', 'ConnectionResolver',
    'MemoryDiscovery', 'DefaultDiscoveryFactory' ]

from .ConnectionParams import ConnectionParams
from .IDiscovery import IDiscovery
from .ConnectionResolver import ConnectionResolver
from .MemoryDiscovery import MemoryDiscovery
from .DefaultDiscoveryFactory import DefaultDiscoveryFactory




# -*- coding: utf-8 -*-
"""
    pip_services3_components.logs.LogMessage
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Log message implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import datetime

class LogMessage(object):
    """
    Data object to store captured log messages. This object is used by [[CachedLogger]].
    """
    time = None
    source = None
    level = None
    correlation_id = None
    error = None
    message = None

    def __init__(self, level = None, source = None, correlation_id = None, error = None, message = None):
        """
        Creates log message

        :param level: a log level.

        :param source: the source (context name)

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param error: an error object associated with this message.

        :param message: a human-readable message to log.
        """
        self.time = datetime.datetime.utcnow()
        self.level = level
        self.source = source
        self.correlation_id = correlation_id
        self.error = error
        self.message = message



# -*- coding: utf-8 -*-
"""
    pip_services3_components.counters.Timing
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Timing implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import time

class Timing:
    """
    Callback object returned by ICounters.beginTiming to end timing
    of execution block and update the associated counter.

    Example:
        timing = counters.begin_timing("mymethod.exec_time")
        ...
        timing.endTiming()
    """

    _start = None
    _callback = None
    _counter = None

    def __init__(self, counter = None, callback = None):
        """
        Creates a new instance of the timing callback object.

        :param counter: an associated counter name

        :param callback: a callback that shall be called when endTiming is called.
        """

        self._counter = counter
        self._callback = callback
        self._start = time.perf_counter() * 1000

    def end_timing(self):
        """
        Ends timing of an execution block, calculates elapsed time and updates the associated counter.
        """

        if not (self._callback is None):
            elapsed = time.perf_counter() * 1000 - self._start
            self._callback.end_timing(self._counter, elapsed)


# -*- coding: utf-8 -*-
"""
    pip_services3_components.log.LogLevel
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Log level enumeration
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

class LogLevel(object):
    """
    Standard log levels.

    Logs at debug and trace levels are usually captured only locally for troubleshooting
    and never sent to consolidated log services.
    """

    Nothing = 0
    """
    Nothing to be logged
    """

    Fatal = 1
    """
    Logs only fatal errors that cause microservice to fail
    """

    Error = 2
    """
    Logs all errors - fatal or recoverable
    """

    Warn = 3
    """
    Logs errors and warnings
    """

    Info = 4
    """
    Logs errors and important information messages
    """

    Debug = 5
    """
    Logs everything up to high-level debugging information
    """

    Trace = 6
    """
    Logs everything down to fine-granular debugging messages
    """



# -*- coding: utf-8 -*-
"""
    pip_services3_components.config.MemoryConfigReader
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Memory config reader implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.config.ConfigParams import ConfigParams
from .IConfigReader import IConfigReader
from pip_services3_commons.config.IReconfigurable import IReconfigurable

class MemoryConfigReader(IConfigReader, IReconfigurable):
    """
    Config reader that stores configuration in memory.

    ### Configuration parameters ###

        The configuration parameters are the configuration template

    Example:
        config = ConfigParams.from_tuples("connection.host", "{{SERVICE_HOST}}",
                                         "connection.port", "{{SERVICE_PORT}}{{^SERVICE_PORT}}8080{{/SERVICE_PORT}}")

        configReader = MemoryConfigReader()
        configReader.configure(config)

        parameters = ConfigParams.fromValue(os.get_env())
        configReader.readConfig("123", parameters)
    """
    _config = None

    def __init__(self, config = None):
        """
        Creates a new instance of config reader.

        :param config: (optional) component configuration parameters
        """
        self._config = config
        
    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        self._config = config

    #todo
    def read_config(self, correlation_id, parameters):
        """
        Reads configuration and parameterize it with given values.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param parameters: values to parameters the configuration or null to skip parameterization.

        :return: ConfigParams configuration.
        """
        return ConfigParams(self._config)

    def read_config_section(self, section):
        config = self._config.get_section(section) if self._not (config is None) else None
        return config



# -*- coding: utf-8 -*-
"""
    pip_services3_components.auth.CredentialParams
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Credential parameters implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.config.ConfigParams import ConfigParams
from pip_services3_commons.data import StringValueMap


class CredentialParams(ConfigParams):
    """
    Contains credentials to authenticate against external services.

    They are used together with connection parameters,
    but usually stored in a separate store, protected from unauthorized access.

    ### Configuration parameters ###

        - store_key:     key to retrieve parameters from credential store
        - username:      user name
        - user:          alternative to username
        - password:      user password
        - pass:          alternative to password
        - access_id:     application access id
        - client_id:     alternative to access_id
        - access_key:    application secret key
        - client_key:    alternative to access_key
        - secret_key:    alternative to access_key

    In addition to standard parameters CredentialParams
    may contain any number of custom parameters

    Example:
        credential = CredentialParams.from_tuples
        ("user", "jdoe", "pass", "pass123", "pin", "321")

        username = credential.get_username()       // Result: "jdoe"
        password = credential.get_password()       // Result: "pass123"
        pin = credential.get_as_nullable_string("pin")// Result: 321
    """

    def __init__(self, values = None):
        """
        Creates a new credential parameters and fills it with values.

        :param values: (optional) an object to be converted into key-value pairs to initialize these credentials.
        """
        super(CredentialParams, self).__init__(values)

    def use_credential_store(self):
        """
        Checks if these credential parameters shall be retrieved from [[CredentialStore]].
        The credential parameters are redirected to [[CredentialStore]] when store_key parameter is set.

        :return: true if credentials shall be retrieved from [[CredentialStore]]
        """
        return "store_key" in self

    def get_store_key(self):
        """
        Gets the key to retrieve these credentials from [[CredentialStore]].
        If this key is null, than all parameters are already present.

        :return: the store key to retrieve credentials.
        """
        return self.get_as_nullable_string("store_key")

    def set_store_key(self, value):
        """
        Sets the key to retrieve these parameters from [[CredentialStore]].

        :param value: a new key to retrieve credentials.
        """
        self.put("store_key", value)

    def get_username(self):
        """
        Gets the user name. The value can be stored in parameters "username" or "user".

        :return: the user name.
        """
        username = self.get_as_nullable_string("username")
        username = username if not (username is None) else self.get_as_nullable_string("user")
        return username

    def set_username(self, value):
        """
        Sets the user name.

        :param value: a new user name.
        """
        self.put("username", value)

    def get_password(self):
        """
        Get the user password. The value can be stored in parameters "password" or "pass".

        :return: the user password.
        """
        password = self.get_as_nullable_string("password")
        password = password if not (password is None) else self.get_as_nullable_string("pass")
        return password

    def set_password(self, password):
        """
        Sets the user password.

        :param password: a new user password.
        """
        self.put("password", password)

    def get_access_id(self):
        """
        Gets the application access id. The value can be stored in parameters "access_id" pr "client_id"

        :return: the application access id.
        """
        access_id = self.get_as_nullable_string("access_id")
        access_id = access_id if not (access_id is None) else self.get_as_nullable_string("client_id")
        return access_id

    def set_access_id(self, value):
        """
        Sets the application access id.

        :param value: a new application access id.
        """
        self.put("access_id", value)

    def get_access_key(self):
        """
        Gets the application secret key.
        The value can be stored in parameters "access_key", "client_key" or "secret_key".

        :return: the application secret key.
        """
        access_key = self.get_as_nullable_string("access_key")
        access_key = access_key if not (access_key is None) else self.get_as_nullable_string("access_key")
        return access_key

    def set_access_key(self, value):
        """
        Sets the application secret key.

        :param value: a new application secret key.
        """
        self.put("access_key", value)

    @staticmethod
    def from_string(line):
        """
        Creates a new CredentialParams object filled with key-value pairs serialized as a string.

        :param line: a string with serialized key-value pairs as "key1=value1;key2=value2;..."
                     Example: "Key1=123;Key2=ABC;Key3=2016-09-16T00:00:00.00Z"

        :return: a new CredentialParams object.
        """
        map = StringValueMap.from_string(line)
        return CredentialParams(map)

    @staticmethod
    def from_tuples(*tuples):
        """
        Creates a new CredentialParams object filled with provided key-value pairs called tuples.
        Tuples parameters contain a sequence of key1, value1, key2, value2, ... pairs.

        :param tuples: the tuples to fill a new CredentialParams object.

        :return: a new CredentialParams object.
        """
        map = StringValueMap.from_tuples_array(tuples)
        return CredentialParams(map)

    @staticmethod
    def many_from_config(config):
        """
        Retrieves all CredentialParams from configuration parameters
        from "credentials" section. If "credential" section is present instead,
        than it returns a list with only one CredentialParams.

        :param config: a configuration parameters to retrieve credentials

        :return: a list of retrieved CredentialParams
        """
        result = []

        # Try to get multiple credentials first
        credentials = config.get_section("credentials")
        if len(credentials) > 0:
            sections_names = credentials.get_section_names()
            for section in sections_names:
                credential = credentials.get_section(section)
                result.append(CredentialParams(credential))
        # Then try to get a single credential
        else:
            credential = config.get_section("credential")
            result.append(CredentialParams(credential))

        return result

    @staticmethod
    def from_config(config):
        """
        Retrieves a single CredentialParams from configuration parameters
        from "credential" section. If "credentials" section is present instead,
        then is returns only the first credential element.

        :param config: ConfigParams, containing a section named "credential(s)".

        :return: the generated CredentialParams object.
        """
        credentials = CredentialParams.many_from_config(config)
        return credentials[0] if len(credentials) > 0 else None



# -*- coding: utf-8 -*-
"""
    pip_services3_components.config.YamlConfigReader
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    YAML config reader implementation

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import os.path
import yaml

from pip_services3_commons.errors.FileException import FileException
from pip_services3_commons.errors.ConfigException import ConfigException
from pip_services3_commons.config.ConfigParams import ConfigParams
from .FileConfigReader import FileConfigReader


class YamlConfigReader(FileConfigReader):
    """
    Config reader that reads configuration from YAML file.

    ### Configuration parameters ###

        - path:          path to configuration file
        - parameters:    this entire section is used as template parameters
        - ...

    Example:
        ======== config.yml ======
        key1: "{{KEY1_VALUE}}"
        key2: "{{KEY2_VALUE}}"
        ===========================

        configReader = YamlConfigReader("config.yml")
        parameters = ConfigParams.from_tuples("KEY1_VALUE", 123, "KEY2_VALUE", "ABC")
        configReader.read_config("123", parameters)
    """

    def __init__(self, path):
        """
        Creates a new instance of the config reader.

        :param path: (optional) a path to configuration file.
        """
        super(YamlConfigReader, self).__init__(path)

    def _read_object(self, correlation_id, parameters):
        """
        Reads configuration file, parameterizes its content and converts it into JSON object.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param parameters: values to parameters the configuration.

        :return: a JSON object with configuration.
        """
        path = self.get_path()

        if path is None:
            raise ConfigException(correlation_id, "NO_PATH", "Missing config file path")

        if not os.path.isfile(path):
            raise FileException(correlation_id, 'FILE_NOT_FOUND', 'Config file was not found at ' + path)

        try:
            with open(path, 'r') as file:
                config = file.read()
                config = self._parameterize(config, parameters)
                return yaml.load(config, Loader=yaml.FullLoader)
        except Exception as ex:
            raise FileException(
                correlation_id,
                "READ_FAILED",
                "Failed reading configuration " + path + ": " + str(ex)
            ).with_details("path", path).with_cause(ex)

    def read_config(self, correlation_id, parameters):
        """
        Reads configuration and parameterize it with given values.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param parameters: values to parameters the configuration or null to skip parameterization.

        :return: ConfigParams configuration.
        """
        value = self._read_object(correlation_id, parameters)
        return ConfigParams.from_value(value)

    @staticmethod
    def read_object(correlation_id, path, parameters):
        """
        Reads configuration file, parameterizes its content and converts it into JSON object.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param path: a path to configuration file.

        :param parameters: values to parameters the configuration.

        :return: a JSON object with configuration.
        """
        return YamlConfigReader(path)._read_object(correlation_id, parameters)

    @staticmethod
    def read_config(correlation_id, path, parameters):
        """
        Reads configuration from a file, parameterize it with given values and returns a new ConfigParams object.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param path: a path to configuration file.

        :param parameters: values to parameters the configuration.

        :return: ConfigParams configuration.
        """
        value = YamlConfigReader(path)._read_object(correlation_id, parameters)
        return ConfigParams.from_value(value)


# -*- coding: utf-8 -*-
"""
    pip_services3_components.log.ConsoleLogger
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Console logger implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import sys
import datetime
import traceback

from .LogLevel import LogLevel
from .ILogger import ILogger
from .LogLevelConverter import LogLevelConverter
from .Logger import Logger


class ConsoleLogger(Logger):
    """
    Logger that writes log messages to console.

    Errors are written to standard err stream and all other messages to standard out stream.

    ### Configuration parameters ###

        - level:             maximum log level to capture
        - source:            source (context) name

    ### References ###
        - *:context-info:*:*:1.0     (optional) [[ContextInfo]] to detect the context id and specify counters source

    Example:
        logger = ConsoleLogger()
        logger.set_level(LogLevel.debug)

        logger.error("123", ex, "Error occured: %s", ex.message)
        logger.debug("123", "Everything is OK.")
    """

    def __init__(self):
        super().__init__()

    def _write(self, level, correlation_id, error, message):
        """
        Writes a log message to the logger destination.

        :param level: a log level.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param error: an error object associated with this message.

        :param message: a human-readable message to log.
        """
        if self._level < level:
            return

        output = "["
        output += correlation_id if correlation_id is not None else "---"
        output += ":"
        output += LogLevelConverter.to_string(level)
        output += ":"
        output += datetime.datetime.utcnow().isoformat()
        output += "] "

        output += message

        if error is not None:
            if len(message) == 0:
                output += "Error: "
            else:
                output += ": "

            output += self._compose_error(error)

        output += "\n"

        if LogLevel.Fatal <= level <= LogLevel.Warn:
            sys.stderr.write(output)
        else:
            sys.stdout.write(output)



# -*- coding: utf-8 -*-
"""
    pip_services3_components.config.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Contains implementation of the config design pattern.

    ConfigReader's Parameterize method allows us to take a standard configuration, and,
    using a set of current parameters (e.g. environment variables), parameterize it. When
    we create the configuration of a container, we can use environment variables to tailor
    it to the system, dynamically add addresses, ports, etc.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = [
    'IConfigReader', 'ConfigReader', 'MemoryConfigReader',
    'FileConfigReader', 'JsonConfigReader', 'YamlConfigReader',
    'DefaultConfigReaderFactory'
]

from .IConfigReader import IConfigReader
from .ConfigReader import ConfigReader
from .MemoryConfigReader import MemoryConfigReader
from .FileConfigReader import FileConfigReader
from .JsonConfigReader import JsonConfigReader
from .YamlConfigReader import YamlConfigReader
from .DefaultConfigReaderFactory import DefaultConfigReaderFactory


# -*- coding: utf-8 -*-
"""
    pip_services3_components.cache.CacheEntry
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Cache entry implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import time

class CacheEntry(object):
    """
    Data object to store cached values with their keys used by [[MemoryCache]]
    """
    expiration = None
    key = None
    value = None

    def __init__(self, key, value, timeout):
        """
        Creates a new instance of the cache entry and assigns its values.

        :param key: a unique key to locate the value.

        :param value: a value to be stored.

        :param timeout: expiration timeout in milliseconds.
        """
        self.key = key
        self.value = value
        self.expiration = time.perf_counter() * 1000 + timeout

    def set_value(self, value, timeout):
        """
        Sets a new value and extends its expiration.

        :param value: a new cached value.

        :param timeout: a expiration timeout in milliseconds.
        """
        self.value = value
        self.expiration = time.perf_counter() * 1000 + timeout

    def is_expired(self):
        """
        Checks if this value already expired.

        :return: true if the value already expires and false otherwise.
        """
        return self.expiration < time.perf_counter() * 1000



# -*- coding: utf-8 -*-
"""
    pip_services3_components.cache.DefaultCacheFactory
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Default cache factory implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .NullCache import NullCache
from .MemoryCache import MemoryCache

from pip_services3_commons.refer.Descriptor import Descriptor
from ..build.Factory import Factory

DefaultCacheFactoryDescriptor = Descriptor(
    "pip-services", "factory", "cache", "default", "1.0"
)

NullCacheDescriptor = Descriptor(
    "pip-services", "cache", "null", "*", "1.0"
)

MemoryCacheDescriptor = Descriptor(
    "pip-services", "cache", "memory", "*", "1.0"
)

class DefaultCacheFactory(Factory):
    """
    Creates [[ICache]] components by their descriptors.
    """
    def __init__(self):
        """
        Create a new instance of the factory.
        """
        self.register_as_type(NullCacheDescriptor, NullCache)
        self.register_as_type(MemoryCacheDescriptor, MemoryCache)



# -*- coding: utf-8 -*-
"""
    pip_services3_components.connect.IDicovery
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Discovery service interface
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .ConnectionParams import ConnectionParams

class IDiscovery:
    """
    Interface for discovery services which are used to store and resolve connection parameters to connect to external services.
    """
    def register(self, correlation_id, key, connection):
        """
        Registers connection parameters into the discovery service.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a key to uniquely identify the connection parameters.

        :param connection: a connection to be registered.
        """
        raise NotImplementedError('Method from interface definition')

    def resolve_one(self, correlation_id, key):
        """
        Resolves a single connection parameters by its key.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a key to uniquely identify the connection.

        :return: a resolved connection.
        """
        raise NotImplementedError('Method from interface definition')

    def resolve_all(self, correlation_id, key):
        """
        Resolves all connection parameters by their key.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a key to uniquely identify the connections.

        :return: a list with resolved connections.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services3_components.counters.LogCounters
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Log counters implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .CachedCounters import CachedCounters
from .CounterType import CounterType
from ..log.CompositeLogger import CompositeLogger
from pip_services3_commons.convert.StringConverter import StringConverter
from pip_services3_commons.refer.IReferenceable import IReferenceable

class LogCounters(CachedCounters, IReferenceable):
    """
    Performance counters that periodically dumps counters measurements to logger.

    ### Configuration parameters ###

        - options:
            - interval:          interval in milliseconds to save current counters measurements (default: 5 mins)
            - reset_timeout:     timeout in milliseconds to reset the counters. 0 disables the reset (default: 0)

    ### References ###
        - *:logger:*:*:1.0           [[ILogger]] components to dump the captured counters
        - *:context-info:*:*:1.0     (optional) [[ContextInfo]] to detect the context id and specify counters source

    Example:
        counters = LogCounters()
        counters.set_references(References.from_tuples(
                    Descriptor("pip-services", "logger", "console", "default", "1.0"), ConsoleLogger()))

        counters.increment("mycomponent.mymethod.calls")
        timing = counters.begin_timing("mycomponent.mymethod.exec_time")
        ...
        timing.endTiming()
    """
    _logger = None

    def __init__(self):
        """
        Creates a new instance of the counters.
        """
        super(LogCounters, self).__init__()
        self._logger = CompositeLogger() 

    #
    # def get_descriptor(self):
    #     return LogCountersDescriptor


    def set_references(self, references):
        """
        Sets references to dependent components.

        :param references: references to locate the component dependencies.
        """
        self._logger.set_references(references)


    def _counter_to_string(self, counter):
        result = "Counter " + counter.name + " { "
        result += "\"type\": " + str(counter.type)
        if not (counter.last is None):
            result += ", \"last\": " + StringConverter.to_string(counter.last)
        if not (counter.count is None):
            result += ", \"count\": " + StringConverter.to_string(counter.count)
        if not (counter.min is None):
            result += ", \"min\": " + StringConverter.to_string(counter.min)
        if not (counter.max is None):
            result += ", \"max\": " + StringConverter.to_string(counter.max)
        if not (counter.average is None):
            result += ", \"avg\": " + StringConverter.to_string(counter.average)
        if not (counter.time is None):
            result += ", \"time\": " + StringConverter.to_string(counter.time)
        result += " }"
        return result

    @staticmethod
    def _get_counter_name(counter):
        return counter.name

    def _save(self, counters):
        """
        Saves the current counters measurements.

        :param counters: current counters measurements to be saves.
        """
        if self._logger is None:
            return
        if len(counters) == 0:
            return

        # Sort counters by name
        counters = sorted(counters, key=LogCounters._get_counter_name)

        for counter in counters:
            self._logger.info("counters", self._counter_to_string(counter))



# -*- coding: utf-8 -*-
"""
    pip_services3_components.cache.MemoryCache
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Memory cache component implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import time
import threading

from .ICache import ICache
from .CacheEntry import CacheEntry
from pip_services3_commons.config.IReconfigurable import IReconfigurable
from pip_services3_commons.run.ICleanable import ICleanable

class MemoryCache(ICache, IReconfigurable, ICleanable):
    """
    Cache that stores values in the process memory.

    Remember: This implementation is not suitable for synchronization of distributed processes.

    ### Configuration parameters ###

    options:
        - timeout:               default caching timeout in milliseconds (default: 1 minute)
        - max_size:              maximum number of values stored in this cache (default: 1000)

    Example:
        cache = MemoryCache()
        cache.store("123", "key1", "ABC", 0)
    """

    _default_timeout = 60000
    _default_max_size = 1000

    _cache = None
    _count = None
    _timeout = None
    _max_size = None
    _lock = None

    def __init__(self):
        """
        Creates a new instance of the cache.
        """
        self._cache = {}
        self._count = 0
        self._max_size = self._default_max_size
        self._timeout = self._default_timeout
        self._lock = threading.Lock()

    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        self._timeout = config.get_as_long_with_default("options.timeout", self._default_timeout)
        self._max_size = config.get_as_long_with_default("options.max_size", self._default_max_size)

    def _cleanup(self):
        oldest = None
        self._count = 0
        
        # Cleanup obsolete entries and find the oldest
        for (key, entry) in self._cache.items():
            # Remove obsolete entry
            if entry.is_expired():
                self._cache.pop(key, None)
            # Count the remaining entry 
            else:
                self._count += 1
                if oldest is None or oldest.expiration > entry.expiration:
                    oldest = entry
        
        # Remove the oldest if cache size exceeded maximum
        if self._count > self._max_size and not (oldest is None):
            self._cache.pop(oldest.key, None)
            self._count -= 1

    def retrieve(self, correlation_id, key):
        """
        Retrieves cached value from the cache using its key.
        If value is missing in the cache or expired it returns None.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a unique value key.

        :return: a cached value or None if value wasn't found or timeout expired.
        """
        self._lock.acquire()
        try:
            # Cache has nothing
            if key not in self._cache:
                return None
                
            # Get entry from the cache
            entry = self._cache[key]
                    
            # Remove entry if expiration set and entry is expired
            if entry.is_expired():
                self._cache.pop(key, None)
                self._count -= 1
                return None
            
            # Update access timeout
            return entry.value
        finally:
            self._lock.release()

    def store(self, correlation_id, key, value, timeout):
        """
        Stores value in the cache with expiration time.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a unique value key.

        :param value: a value to store.

        :param timeout: expiration timeout in milliseconds.

        :return: a cached value stored in the cache.
        """
        timeout = timeout if timeout > 0 else self._default_timeout

        self._lock.acquire()
        try:
            entry = None
            if key in self._cache:
                entry = self._cache[key]

            # Shortcut to remove entry from the cache
            if value is None:
                if not (entry is None):
                    self._cache.pop(key, None)
                    self._count -= 1
                return None
            
            # Update the entry
            if not (entry is None):
                entry.set_value(value, timeout)
            # Or create a new entry 
            else:
                entry = CacheEntry(key, value, timeout)
                self._cache[key] = entry
                self._count += 1

            # Clean up the cache
            if self._max_size > 0 and self._count > self._max_size:
                self._cleanup()
            
            return value
        finally:
            self._lock.release()
    
    def remove(self, correlation_id, key):
        """
        Removes a value from the cache by its key.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a unique value key.
        """
        self._lock.acquire()
        try:
            # Get the entry
            entry = self._cache.pop(key, None)

            # Remove entry from the cache
            if not (entry is None):
                self._count -= 1
        finally:
            self._lock.release()
    
    def clear(self, correlation_id):
        """
        Clears component state.

        :param correlation_id: (optional) transaction id to trace execution through call chain.
        """
        self._lock.acquire()
        try:
            self._cache = {}
        finally:
            self._lock.release()
    


# -*- coding: utf-8 -*-
"""
    pip_services3_components.connect.MemoryDiscovery
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Memory discovery implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.config.ConfigParams import ConfigParams
from pip_services3_commons.config.IReconfigurable import IReconfigurable
from .IDiscovery import IDiscovery
from .ConnectionParams import ConnectionParams

class DiscoveryItem:
    """
    Used to store key-identifiable information about connections.
    """
    key = None
    connection = None

class MemoryDiscovery(IDiscovery, IReconfigurable):
    """
    Discovery service that keeps connections in memory.

    ### Configuration parameters ###

        - [connection key 1]:
        - ...                          connection parameters for key 1
        - [connection key 2]:
        - ...                          connection parameters for key N

    Example:
        config = ConfigParams.from_tuples(
        "key1.host", "10.1.1.100",
        "key1.port", "8080",
        "key2.host", "10.1.1.100",
        "key2.port", "8082")

        discovery = MemoryDiscovery()
        discovery.read_connections(config)

        discovery.resolve("123", "key1")
    """
    _items = None

    def __init__(self, config = None):
        """
        Creates a new instance of discovery service.

        :param config: (optional) configuration with connection parameters.
        """
        self._items = []
        if not (config is None):
            self.configure(config)

    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        self.read_connections(config)

    def read_connections(self, connections):
        """
        Reads connections from configuration parameters.
        Each section represents an individual Connectionparams

        :param connections: configuration parameters to be read
        """
        del self._items[:]
        for key in connections.get_key_names():
            item = DiscoveryItem()
            item.key = key
            value = connections.get_as_nullable_string(key)
            item.connection = ConnectionParams.from_string(value)
            self._items.append(item)

    def register(self, correlation_id, key, connection):
        """
        Registers connection parameters into the discovery service.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a key to uniquely identify the connection parameters.

        :param connection: a connection to be registered.
        """
        item = DiscoveryItem()
        item.key = key
        item.connection = connection
        self._items.append(item)

    def resolve_one(self, correlation_id, key):
        """
        Resolves a single connection parameters by its key.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a key to uniquely identify the connection.

        :return: a resolved connection.
        """
        connection = None
        for item in self._items:
            if item.key == key and not (item.connection is None):
                connection = item.connection
                break
        return connection

    def resolve_all(self, correlationId, key):
        """
        Resolves all connection parameters by their key.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param key: a key to uniquely identify the connections.

        :return: a list with resolved connections.
        """
        connections = []
        for item in self._items:
            if item.key == key and not (item.connection is None):
                connections.append(item.connection)
        return connections



# -*- coding: utf-8 -*-
"""
    pip_services3_components.connect.ConnectionResolver
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Connection resolver implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from pip_services3_commons.config import ConfigParams

from .ConnectionParams import ConnectionParams
from .IDiscovery import IDiscovery
from pip_services3_commons.config.IConfigurable import IConfigurable
from pip_services3_commons.refer.IReferenceable import IReferenceable
from pip_services3_commons.refer.Descriptor import Descriptor
from pip_services3_commons.errors.ConfigException import ConfigException

class ConnectionResolver(IConfigurable, IReferenceable):
    """
    Helper class to retrieve component connections.

    If connections are configured to be retrieved from [[IDiscovery]],
    it automatically locates [[IDiscovery]] in component references
    and retrieve connections from there using discovery_key parameter.

    ### Configuration parameters ###

        - connection:
            - discovery_key:               (optional) a key to retrieve the connection from IDiscovery
            - ...                          other connection parameters

        - connections:                  alternative to connection
            - [connection params 1]:       first connection parameters
            - ...                      connection parameters for key 1
            - [connection params N]:       Nth connection parameters
            - ...                      connection parameters for key N

    ### References ###
        - *:discovery:*:*:1.0    (optional) IDiscovery services to resolve connections

    Example:
        config = ConfigParams.from_tuples("connection.host", "10.1.1.100", "connection.port", 8080)

        connectionResolver = ConnectionResolver()
        connectionResolver.configure(config)
        connectionResolver.setReferences(references)
        connectionResolver.resolve("123")
    """
    _connections = None
    _references = None

    def __init__(self, config = None, references = None):
        """
        Creates a new instance of connection resolver.

        :param config: (optional) component configuration parameters

        :param references: (optional) component references
        """
        self._connections = []
        if not (config is None):
            self.configure(config)
        if not (references is None):
            self.set_references(references)

    def set_references(self, references):
        """
        Sets references to dependent components.

        :param references: references to locate the component dependencies.
        """
        self._references = references

    def configure(self, config):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        connections = ConnectionParams.many_from_config(config)
        for connection in connections:
            self._connections.append(connection)

    def get_all(self):
        """
        Gets all connections configured in component configuration.
        Redirect to Discovery services is not done at this point.
        If you need fully fleshed connection use [[resolve]] method instead.

        :return: a list with connection parameters
        """
        return self._connections

    def add(self, connection):
        """
        Adds a new connection to component connections

        :param connection: new connection parameters to be added
        """
        self._connections.append(connection)

    def _register_in_discovery(self, correlation_id, connection):
        if connection.use_discovery() == False: return False
        
        key = connection.get_discovery_key()
        if self._references is None:
            return False
        
        descriptor = Descriptor("*", "discovery", "*", "*", "*")
        components = self._references.get_optional(descriptor)
        if components is None:
            return False
        
        for component in components:
            if isinstance(component, IDiscovery):
                component.register(correlation_id, key, connection)
        
        return True

    def register(self, correlation_id, connection):
        """
        Registers the given connection in all referenced discovery services.
        This method can be used for dynamic service discovery.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param connection: a connection to register.
        """
        result = self._register_in_discovery(correlation_id, connection)
        
        if result:
            self._connections.append(connection)

    def _resolve_in_discovery(self, correlation_id, connection):
        if connection.use_discovery() == False: return None
        
        key = connection.get_discovery_key()
        descriptor = Descriptor("*", "discovery", "*", "*", "*")
        components = self._references.get_optional(descriptor)
        if len(components) == 0:
            raise ConfigException(correlation_id, "CANNOT_RESOLVE", "Discovery wasn't found to make resolution")

        for component in components:
            if isinstance(component, IDiscovery):
                resolved_connection = component.resolve_one(correlation_id, key)
                if not (resolved_connection is None):
                    return resolved_connection
        
        return None

    def resolve(self, correlation_id):
        """
        Resolves a single component connection. If connections are configured to be retrieved
        from Discovery service it finds a [[IDiscovery]] and resolves the connection there.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :return: resolved connection parameters or null if nothing was found.
        """
        if len(self._connections) == 0: return None
        
        # Return connection that doesn't require discovery
        for connection in self._connections:
            if not connection.use_discovery():
                return connection
        
        # Return connection that require discovery
        for connection in self._connections:
            if connection.use_discovery():
                resolved_connection = self._resolve_in_discovery(correlation_id, connection)
                if not (resolved_connection is None):
                    # Merge configured and new parameters
                    resolved_connection = ConnectionParams(ConfigParams.merge_configs(connection, resolved_connection))
                    return resolved_connection
        
        return None

    def _resolve_all_in_discovery(self, correlation_id, connection):
        result = []
        
        if connection.use_discovery() == False: return result
        
        key = connection.get_discovery_key()
        descriptor = Descriptor("*", "discovery", "*", "*", "*")
        components = self._references.get_optional(descriptor)
        if len(components) == 0:
            raise ConfigException(correlation_id, "CANNOT_RESOLVE", "Discovery wasn't found to make resolution")

        for component in components:
            if isinstance(component, IDiscovery):
                resolved_connections = component.resolve_all(correlation_id, key)
                if not (resolved_connections is None):
                    for connection in resolved_connections:
                        result.append(connection)
        
        return result

    def resolve_all(self, correlation_id):
        """
        Resolves all component connection. If connections are configured to be retrieved
        from Discovery service it finds a [[IDiscovery]] and resolves the connection there.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :return: a list of resolved connections.
        """
        resolved = []
        to_resolve = []

        # Sort connections
        for connection in self._connections:
            if connection.use_discovery():
                to_resolve.append(connection)
            else:
                resolved.append(connection)
        
        # Resolve addresses that require that
        if len(to_resolve) > 0:
            for connection in to_resolve:
                resolved_connections = self._resolve_all_in_discovery(correlation_id, connection)
                for resolved_connection in resolved_connections:
                    # Merge configured and new parameters
                    resolved_connection = ConnectionParams(ConfigParams.merge_configs(connection, resolved_connection))
                    resolved.append(resolved_connection)
        
        return resolved



# -*- coding: utf-8 -*-
"""
    pip_services3_components.log.CompositeLogger
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Composite logger implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .ILogger import ILogger
from .Logger import Logger
from pip_services3_commons.refer.Descriptor import Descriptor
from pip_services3_commons.refer.IReferenceable import IReferenceable

class CompositeLogger(Logger, IReferenceable):
    """
    Aggregates all loggers from component references under a single component.

    It allows to log messages and conveniently send them to multiple destinations.

    ### References ###
        - *:logger:*:*:1.0 	(optional) [[ILogger]] components to pass log messages

    Example:
        class MyComponent(IConfigurable, IReferenceable):
            _logger = CompositeLogger()

            def configure(self, config):
                self._logger.configure(config)

            def set_references(self, references):
                self._logger.set_references(references)

            def my_method(self, correlation_id):
                self._logger.debug(correlationId, "Called method mycomponent.mymethod")

            ...
    """
    _loggers = None

    def __init__(self, references = None):
        """
        Creates a new instance of the logger.

        :param references: references to locate the component dependencies.
        """
        self._loggers = []

        if not (references is None):
            self.set_references(references)
            
    def set_references(self, references):
        """
        Sets references to dependent components.

        :param references: references to locate the component dependencies.
        """
        super(CompositeLogger, self).set_references(references)
        # descriptor = Descriptor(None, "logger", None, None, None)
        loggers = references.get_optional(Descriptor(None, "logger", None, None, None))
        for logger in loggers:
            if isinstance(logger, ILogger):
                self._loggers.append(logger)

    def _write(self, level, correlation_id, error, message):
        """
        Writes a log message to the logger destination.

        :param level: a log level.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param error: an error object associated with this message.

        :param message: a human-readable message to log.
        """
        for logger in self._loggers:
            logger.log(level, correlation_id, error, message)




# -*- coding: utf-8 -*-
"""
    pip_services3_components.build.CreateException
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Create exception type
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from pip_services3_commons.errors.InternalException import InternalException
from pip_services3_commons.refer import ReferenceException


class CreateException(InternalException):
    """
    Error raised when factory is not able to create requested component.
    """
    def __init__(self, correlation_id = None, message = None):
        """
        Creates an error instance and assigns its values.

        :param correlation_id: (optional) a unique transaction id to trace execution through call chain.

        :param message: human-readable error or locator of the component that cannot be created.
        """
        super(CreateException, self).__init__(correlation_id, "CANNOT_CREATE", message)



# -*- coding: utf-8 -*-
"""
    pip_services3_components.count.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Performance counters. They show non-functional characteristics about how the code works,
    like: times called, response time, objects saved/processed. Using these numbers, we can
    show how the code works in the system – how stable, fast, expandable it is.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = [
    'CounterType', 'ITimingCallback', 'ICounters',
    'Counter', 'Timing', 'CachedCounters', 
    'NullCounters', 'CompositeCounters', 'LogCounters',
    'DefaultCountersFactory'
]

from .CounterType import CounterType
from .ITimingCallback import ITimingCallback
from .ICounters import ICounters
from .Counter import Counter
from .Timing import Timing
from .CachedCounters import CachedCounters
from .NullCounters import NullCounters
from .CompositeCounters import CompositeCounters
from .LogCounters import LogCounters
from .DefaultCountersFactory import DefaultCountersFactory



# -*- coding: utf-8 -*-
"""
    pip_services3_components.log.LogLevelConverter
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Log level converter implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .LogLevel import LogLevel

class LogLevelConverter(object):
    """
    Helper class to convert log level values.
    """
    @staticmethod
    def to_log_level(value):
        """
        Converts numbers and strings to standard log level values.

        :param value: a value to be converted

        :return: converted log level
        """
        if value is None:
            return LogLevel.Info

        value = str(value).upper()
        if ("0" == value) or ("NOTHING" == value) or ("NONE" == value):
            return LogLevel.Nothing
        elif ("1" == value) or ("FATAL" == value):
            return LogLevel.Fatal
        elif ("2" == value) or ("ERROR" == value):
            return LogLevel.Error
        elif ("3" == value) or ("WARN" == value) or ("WARNING" == value):
            return LogLevel.Warn
        elif ("4" == value) or ("INFO" == value):
            return LogLevel.Info
        elif ("5" == value) or ("DEBUG" == value):
            return LogLevel.Debug
        elif ("6" == value) or ("TRACE" == value):
            return LogLevel.Trace
        else:
            return LogLevel.Info

    @staticmethod
    def to_string(level):
        """
        Converts log level to a string.

        :param level: a log level to convert

        :return: log level name string.
        """
        if level == LogLevel.Fatal:
            return "FATAL" 
        if level == LogLevel.Error:
            return "ERROR" 
        if level == LogLevel.Warn:
            return "WARN" 
        if level == LogLevel.Info:
            return "INFO" 
        if level == LogLevel.Debug:
            return "DEBUG" 
        if level == LogLevel.Trace:
            return "TRACE"
        return "UNDEF"

    @staticmethod
    def to_integer(level):
        """
        Converts log level to a number.

        :param level: a log level to convert.

        :return: log level number value.
        """
        if level == LogLevel.Fatal:
            return 1 
        if level == LogLevel.Error:
            return 2 
        if level == LogLevel.Warn:
            return 3 
        if level == LogLevel.Info:
            return 4 
        if level == LogLevel.Debug:
            return 5 
        if level == LogLevel.Trace:
            return 6
        return 0




# -*- coding: utf-8 -*-
"""
    pip_services3_components.log.ILogger
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for logging components.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

class ILogger:
    """
    Interface for logger components that capture execution log messages.
    """

    def get_level(self):
        """
        Gets the maximum log level. Messages with higher log level are filtered out.

        :return: the maximum log level.
        """
        raise NotImplementedError('Method from interface definition')

    def set_level(self, level):
        """
        Set the maximum log level.

        :param level: a new maximum log level.
        """
        raise NotImplementedError('Method from interface definition')

    def log(self, level, correlation_id, error, message, *args, **kwargs):
        """
        Logs a message at specified log level.

        :param level: a log level.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param error: an error object associated with this message.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        raise NotImplementedError('Method from interface definition')

    def fatal(self, correlation_id, error, message, *args, **kwargs):
        """
        Logs fatal (unrecoverable) message that caused the process to crash.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param error: an error object associated with this message.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        raise NotImplementedError('Method from interface definition')

    def error(self, correlation_id, error, message, *args, **kwargs):
        """
        Logs recoverable application error.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param error: an error object associated with this message.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        raise NotImplementedError('Method from interface definition')

    def warn(self, correlation_id, message, *args, **kwargs):
        """
        Logs a warning that may or may not have a negative impact.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        raise NotImplementedError('Method from interface definition')

    def info(self, correlation_id, message, *args, **kwargs):
        """
        Logs an important information message

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        raise NotImplementedError('Method from interface definition')

    def debug(self, correlation_id, message, *args, **kwargs):
        """
        Logs a high-level debug information for troubleshooting.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        raise NotImplementedError('Method from interface definition')

    def trace(self, correlation_id, message, *args, **kwargs):
        """
        Logs a low-level debug information for troubleshooting.

        :param correlation_id: (optional) transaction id to trace execution through call chain.

        :param message: a human-readable message to log.

        :param args: arguments to parameterize the message.

        :param kwargs: arguments to parameterize the message.
        """
        raise NotImplementedError('Method from interface definition')

